# libraries
from datetime import datetime, timedelta

import numpy as np
import pandas as pd

from tank_forecaster.decomp import generic_hh_seasonality


def format_tank(dict_input, freq: str = "30min"):

    if dict_input == [] or dict_input is None:
        return None

    else:
        dict_input = sorted(dict_input, key=lambda x: x["read_time"])
        dict_input = [d for d in dict_input if d["volume"]]
        df = pd.DataFrame(dict_input)
        df.read_time = pd.to_datetime(df.read_time)
        df["y"] = df.volume.diff() * -1
        df = df[["read_time", "y"]].rename(columns={"read_time": "ds"})
        df = df[df.y < 800].copy()
        df.y = df.y.clip(lower=0)
        df = df.set_index("ds").groupby(pd.Grouper(freq=freq)).sum()

        if len(df) == 2:
            return None

        df.reset_index(level=0, inplace=True)
        output = df.iloc[1:-1]

        return output


def format_sales(dict_input):

    if dict_input == [] or dict_input is None:
        return None

    else:
        dict_input = sorted(dict_input, key=lambda x: x["date"])
        dat = pd.DataFrame(dict_input)

        # only same product type
        if "tank_type" in dat:
            df = dat[dat["tank_type"] == dat.tank_type.iloc[-1]]
        else:
            df = dat

        df = df[["date", "sales"]]
        df.sales = df.sales.clip(
            lower=df.sales.quantile(0.015), upper=df.sales.quantile(0.995)
        )
        df.rename(columns={"date": "ds", "sales": "y"}, inplace=True)
        df["ds"] = pd.to_datetime(df["ds"], format="%Y-%m-%d")

    return df


def gen_past(df, freq="30min", periods=432):

    if len(df) == 0 or df is None:
        now = datetime.now()
        end_0 = now - (now - datetime.min) % timedelta(minutes=30)

    else:
        end_0 = df.iloc[-1, 0]

    past = pd.date_range(end=end_0, freq=freq, periods=periods)
    past = pd.DataFrame(past)
    past.rename(columns={0: "ds"}, inplace=True)
    past["ds"] = pd.to_datetime(past["ds"], format="%Y-%m-%d")

    return past


def validate_tank(tank_data, sales_data):

    if len(tank_data) == 0 or tank_data is None:
        return None

    # add day of year for tank data / sales data
    sales_data["doy"] = sales_data["ds"].dt.dayofyear
    sales_data = sales_data.iloc[-100:]

    # create validation dataframe
    val_matrix = np.zeros((432, 2))
    val_df = pd.DataFrame(val_matrix, columns=["ds", "doy"])

    # dynamically generate timestamps
    val_df["ds"] = gen_past(tank_data, freq="30min", periods=432)
    val_df["doy"] = val_df.ds.dt.dayofyear
    val_df["ts"] = val_df.ds.dt.time

    val_df = pd.merge(
        left=val_df, right=generic_daily_decomp, how="left", left_on="ts", right_on="ts"
    )
    val_df = pd.merge(
        left=val_df, right=tank_data, how="left", left_on="ds", right_on="ds"
    )
    val_df = pd.merge(
        left=val_df, right=sales_data, how="left", left_on="doy", right_on="doy"
    )

    val_df.rename(columns={"ds_x": "ds", "y_x": "actual", "y_y": "daily"}, inplace=True)
    val_df = val_df.loc[:, ["ds", "ts", "actual", "daily", "daily_multi"]]

    # calculate generalized daily tank sticking curve from sales data
    val_df["base_est"] = val_df.loc[:, "daily"] * (1 / 48)
    est = val_df["base_est"] * val_df["daily_multi"]
    val_df["estimated"] = est

    val_df = val_df.loc[:, ["ds", "actual", "estimated"]]

    # data abnormalities adjustment
    diff = val_df.estimated - val_df.actual
    threshold = diff.abs().quantile(0.95)

    mask_adj = diff.abs() > threshold
    mask_tank = [not i for i in mask_adj]

    # val_df.loc[mask_adj, 'output'] = val_df.loc[mask_adj, 'adj_sales']
    # val_df.loc[mask_tank, 'output'] = val_df.loc[mask_tank, 'tank_vals']

    # missing data imputation
    mask_nan = val_df.actual.isna()
    mask_not_nan = [not i for i in mask_nan]

    val_df.loc[mask_nan, "y"] = val_df.loc[mask_nan, "estimated"]
    val_df.loc[mask_not_nan, "y"] = val_df.loc[mask_not_nan, "actual"]

    # for testing purposes, uncomment this
    # return val_df

    # for use in forecasting, uncomment this
    val_df.rename(columns={"ds_x": "ds"}, inplace=True)

    return val_df[["ds", "y"]]



from datetime import datetime, timedelta

import numpy as np
import pandas as pd
from fbprophet import Prophet

from tank_forecaster.decomp import (
    generic_hh_seasonality,
    generic_weekly_seasonality,
    generic_yearly_seasonality,
)


def gen_future(df, periods=30, freq="1D"):
    start_0 = df.iloc[-1, 0]
    future = pd.date_range(start=start_0, freq=freq, periods=periods) + pd.Timedelta(
        freq
    )
    future = pd.DataFrame(future)
    future.rename(columns={0: "ds"}, inplace=True)

    return future


def forecast_far(
    sales_history,
    yearly_decomp=None,
    weekly_decomp=None,
    forecast_length=90,
    daily_lift_est=0,
):
    # if no decompositions are created, or not a full year of sales data, use generic
    if yearly_decomp is None or len(yearly_decomp) <= 52:
        yearly_decomp = generic_yearly_decomp

    if weekly_decomp is None or len(weekly_decomp) <= 6:
        weekly_decomp = generic_weekly_decomp

    # generate predictions DataFrame
    predictions = pd.DataFrame(columns=["woy", "dow", "base", "weekly", "daily"])
    predictions["woy"] = np.repeat(range(1, 54), 7)  # 52.14 'week of year' -> 53
    predictions["dow"] = [0, 1, 2, 3, 4, 5, 6] * 53  # 7 'day of week' each week of year

    # if no sales data is given, use daily_lift_estimate input w/ generic seasonality
    if sales_history is None or len(sales_history) == 0:
        predictions["base"] = daily_lift_est
        weekly = np.repeat(predictions.base.iloc[1] * 7 * generic_yearly_decomp, 7)
        weekly.index = range(0, 371)
        predictions["weekly"] = weekly
        weekly_decomp_rep = pd.concat([generic_weekly_decomp] * 53, ignore_index=True)
        predictions["daily"] = predictions.weekly * (1 / 7) * weekly_decomp_rep
        data = {
            "ds": [pd.to_datetime(datetime.now()), pd.to_datetime(datetime.now())],
            "sales": [0, 0],
        }
        sales_history = pd.DataFrame(data=data)
        sales_history["date"] = sales_history["ds"].dt.date

    else:  # if full year or more of sales data, use last year's average
        predictions["base"] = sales_history.y[-365:].mean()

    weekly = np.repeat(predictions.base[1] * 7 * yearly_decomp, 7)
    weekly.index = range(0, 371)
    predictions["weekly"] = weekly
    weekly_decomp_rep = pd.concat([weekly_decomp] * 53, ignore_index=True)

    predictions["daily"] = predictions.weekly * (1 / 7) * weekly_decomp_rep

    # match predictions DataFrame to gen future DataFrame
    future = gen_future(sales_history, periods=forecast_length, freq="1D")
    future["dow"] = future.ds.dt.weekday
    future["woy"] = future.ds.dt.week  # deprecated?
    output = pd.merge(
        future, predictions, left_on=["woy", "dow"], right_on=["woy", "dow"]
    )
    output = output[["ds", "daily"]]
    output.rename(columns={"daily": "yhat"}, inplace=True)

    # confidence interval
    output["lower"] = output["yhat"] - 2 * output["yhat"].std()
    output["upper"] = output["yhat"] + 2 * output["yhat"].std()

    # non-negative predictions
    for field in ["yhat", "lower", "upper"]:
        output.loc[output[field] < 0, field] = 0

    return output


def forecast_near(
    validated_tank_history,
    forecast_freq="30min",
    forecast_length=144,
    daily_lift_est=0,
):

    # less than 9 days of historical data
    if validated_tank_history is None:
        validated_tank_history = pd.DataFrame(columns=["ds", "y"])

    if len(validated_tank_history) < 432:

        if validated_tank_history.empty:
            now = datetime.now()
            now_hh = now - (now - datetime.min) % timedelta(minutes=30)

        else:
            now_hh = validated_tank_history["ds"].iloc[-1]

        forecast = forecast_far(
            sales_history=[],
            yearly_decomp=generic_yearly_decomp,
            weekly_decomp=generic_weekly_decomp,
            daily_lift_est=daily_lift_est,
            forecast_length=3,
        )
        fake_data = {"ds": [now_hh], "sales": [0]}
        validated_tank_history = pd.DataFrame(data=fake_data)

        future = gen_future(
            validated_tank_history, periods=forecast_length, freq=forecast_freq
        )

        predictions = pd.DataFrame(columns=["ds", "daily"])
        predictions["ds"] = future.ds
        predictions["daily"] = np.repeat(forecast.yhat[0], len(predictions))
        predictions["ts"] = predictions["ds"].dt.time

        predictions = pd.merge(
            left=predictions,
            right=generic_daily_decomp,
            how="left",
            left_on="ts",
            right_on="ts",
        )

        predictions["hh"] = predictions.daily * (predictions.daily_multi * 1 / 48)
        output = predictions[["ds", "hh"]].copy()
        output.rename(columns={"hh": "yhat"}, inplace=True)

        output["lower"] = output["yhat"] - 2 * output["yhat"].std()
        output["upper"] = output["yhat"] + 2 * output["yhat"].std()

        # non-negative predictions
        for field in ["yhat", "lower", "upper"]:
            output.loc[output[field] < 0, field] = 0

        return output

    future = gen_future(
        validated_tank_history, periods=forecast_length, freq=forecast_freq
    )

    m = Prophet(
        changepoint_prior_scale=0.05,
        daily_seasonality=True,
        weekly_seasonality=True,
        yearly_seasonality=False,
    )

    m.fit(validated_tank_history)

    forecast = m.predict(future)

    # format output
    output = forecast.loc[:, ["ds", "yhat_lower", "yhat_upper", "yhat"]]
    output.rename(columns={"yhat_lower": "lower", "yhat_upper": "upper"}, inplace=True)

    # non-negative predictions
    for field in ["yhat", "lower", "upper"]:
        output.loc[output[field] < 0, field] = 0

    return output



import numpy as np
import pandas as pd
import pytest

from tank_forecaster import forecaster_advanced
from tank_forecaster.decomp import generic_hh_seasonality


def is_a_valid_forecast_dataframe(df):
    return (
        type(df) is pd.core.frame.DataFrame
        and len(df) == 144
        and "ds" in df
        and "yhat" in df
        and type(df.ds.iloc[-1]) is pd._libs.tslibs.timestamps.Timestamp
        and type(df.yhat.iloc[-1]) is np.float64
    )


# @pytest.mark.parameterize(
#     "daily_lift_est,hh_seasonality", [(1000, generic_hh_seasonality), (None, None)]
# )
# def test_near_approx_returns_a_valid_dataframe(daily_lift_est, hh_seasonality):


def test_near_approx_returns_a_valid_dataframe():
    x = forecaster_advanced.forecast_near_approx(
        daily_lift_est=100, hh_seasonality=generic_hh_seasonality
    )
    assert is_a_valid_forecast_dataframe(x)


def test_near_approx_returns_a_valid_dataframe_for_generic_input():
    x = forecaster_advanced.forecast_near_approx(
        daily_lift_est=1000, hh_seasonality=generic_hh_seasonality
    )
    assert is_a_valid_forecast_dataframe(x)


def test_near_returns_none_when_no_data():
    assert forecaster_advanced.forecast_near(None) is None


def test_near_full_data_returns_a_valid_dataframe(tank_full_data):
    x = forecaster_advanced.forecast_near(tank_history=tank_full_data)
    assert is_a_valid_forecast_dataframe(x)


def test_near_approx_start_date():
    pass


def test_near_start_date():
    pass



import numpy as np
import pandas as pd
from fbprophet import Prophet


def decompose_sales(df):

    if df is None:
        return generic_yearly_seasonality, generic_weekly_seasonality

    if len(df) == 0:
        return generic_yearly_seasonality, generic_weekly_seasonality

    else:
        m = Prophet(
            changepoint_prior_scale=0.05,
            daily_seasonality=False,
            weekly_seasonality=True,
            yearly_seasonality=True,
            seasonality_mode="multiplicative",
        )

        m.fit(df)
        future = m.make_future_dataframe(periods=1, freq="1D")
        forecast = m.predict(future)
        forecast_reduced = forecast.loc[:, ["ds", "yearly", "weekly"]]
        forecast_reduced["dow"] = forecast_reduced.ds.dt.weekday
        forecast_reduced["doy"] = forecast_reduced.ds.dt.dayofyear
        forecast_reduced["woy"] = forecast_reduced.ds.dt.week  # deprecated?
        week_trend = forecast_reduced.groupby("dow")["weekly"].mean() + 1
        forecast_reduced = forecast_reduced.iloc[-366:]
        forecast_reduced.sort_values(by=["woy", "dow"], inplace=True)
        year_trend = forecast_reduced.groupby("woy").mean()
        year_trend = year_trend["yearly"] + 1

    return year_trend, week_trend


# tank history decomposition for daily trends
def decompose_tank(tank_history):

    m = Prophet(
        changepoint_prior_scale=0.05,
        daily_seasonality=True,
        weekly_seasonality=True,
        yearly_seasonality=False,
        seasonality_mode="multiplicative",
    )

    m.fit(tank_history)
    future = m.make_future_dataframe(periods=1, freq="30min")
    forecast = m.predict(future)
    forecast_reduced = forecast.loc[:, ["ds", "daily", "daily_lower", "daily_upper"]]
    forecast_reduced["time"] = forecast_reduced["ds"].dt.time
    forecast_reduced.drop(columns="ds", inplace=True)
    day_trend = forecast_reduced.groupby("time").mean()
    day_trend.sort_values(by=["time"], inplace=True)

    for field in ["daily", "daily_lower", "daily_upper"]:
        day_trend.loc[:, field] = day_trend.loc[:, field] + 1

    return day_trend


# holiday decomposition - needs improvement for consistent date mapping
def decompose_holiday(input):
    if type(input) != pd.core.frame.DataFrame:
        return print("error - input is not a DataFrame")
    else:
        event_1 = pd.DataFrame(
            {
                "holiday": "event_1_name",
                "ds": pd.to_datetime(["2020-03-01"]),
                "lower_window": 0,
                "upper_window": 0,
            }
        )
        event_2 = pd.DataFrame(
            {
                "holiday": "event_2_name",
                "ds": pd.to_datetime(["2020-03-02"]),
                "lower_window": 0,
                "upper_window": 0,
            }
        )
        holidays = pd.concat([event_1, event_2])
        m = Prophet(
            holidays=holidays,
            changepoint_prior_scale=0.05,
            daily_seasonality=False,
            weekly_seasonality=True,
            yearly_seasonality=True,
            seasonality_mode="multiplicative",
        )
        m.add_country_holidays(country_name="US")
        m.fit(input)
        future = m.make_future_dataframe(periods=1, freq="1D")
        forecast = m.predict(future)
        forecast_reduced = forecast.loc[:, ["ds", "holidays"]]

    return forecast_reduced


generic_weekly_seasonality = [0.95, 1.0, 1.05, 1.05, 1.2, 0.95, 0.8]

generic_yearly_seasonality = [
    0.95,
    0.95,
    *np.repeat(0.9, 10),
    0.925,
    0.95,
    0.975,
    *np.repeat(1, 7),
    1.025,
    1.05,
    *np.repeat(1.075, 11),
    1.05,
    1.025,
    *np.repeat(1, 7),
    *np.repeat(0.95, 9),
]

generic_hh_seasonality = [
    1.0,
    0.85,
    0.7,
    0.55,
    0.4,
    0.25,
    0.1,
    0.0,
    0.0,
    0.0,
    0.0,
    0.0,
    0.0,
    0.0,
    0.0,
    0.0,
    0.1,
    0.2,
    0.3,
    0.4,
    0.5,
    0.6,
    0.7,
    0.8,
    0.9,
    1.0,
    1.1,
    1.2,
    1.3,
    1.4,
    1.5,
    1.6,
    1.7,
    1.8,
    1.9,
    2.0,
    2.0,
    2.0,
    2.0,
    1.9,
    1.8,
    1.7,
    1.6,
    1.5,
    1.4,
    1.3,
    1.2,
    1.1,
]



import numpy as np
import pandas as pd

from tank_forecaster import forecaster_advanced
from tank_forecaster.decomp import (generic_weekly_seasonality,
                                    generic_yearly_seasonality)


def test_far():
    x = forecaster_advanced.forecast_far(
        daily_lift_est=5000,
        yearly_seasonality=generic_yearly_seasonality,
        weekly_seasonality=generic_weekly_seasonality,
    )
    assert type(x) is pd.core.frame.DataFrame
    assert "ds" in x and "yhat" in x
    assert len(x) == 90
    assert type(x.ds.iloc[-1]) is pd._libs.tslibs.timestamps.Timestamp
    assert type(x.yhat.iloc[-1]) is np.float64


def test_far_start_date():
    x = forecaster_advanced.forecast_far(
        daily_lift_est=5000,
        yearly_seasonality=generic_yearly_seasonality,
        weekly_seasonality=generic_weekly_seasonality,
        start_date="2020-01-01",
    )
    assert str(x.ds.iloc[0]) == "2020-01-01 00:00:00"
    assert len(x) == 90
    assert str(x.ds.iloc[-1]) == "2020-03-30 00:00:00"


if __name__ == "__main__":
    pass
